#lang racket

;;; LAScheme -- or, AspectScheme with Execution Levels
;;;  by Éric Tanter
;;; 
;;; AspectScheme was originally developed by Christopher Dutchyn.
;;; This is an extension of AspectScheme that supports execution levels.
;;; (see "Execution Levels for Aspect-Oriented Programming", by Tanter, AOSD 2010)
;;; main differences with AspectScheme:
;;; - execution of pointcuts and advices happens at a higher level
;;; - no pc/adv join points (not needed anymore, thanks to levels)
;;; - app/prim is not exposed anymore (not needed, thanks to levels)
;;; - exports level-shifting operators ^up and ^down
;;; - exports lambda-s to define level-capturing functions
;;; - changed the interface of the language for creating and deploying aspects
;;; - fixed minor details and bugs, and renamed some functions

;;; This module is the whole LAScheme definition, except that weaving is disabled.
;;; This means that a client module can import it and use LAScheme to program,
;;; but that module is not subject to weaving. This is important to use LAScheme
;;; to design language extensions whose internal behavior shouldn't be woven.

(require (only-in swindle/base let let* letrec))

(provide (all-from-out racket)
         (rename-out [lambda/aspects/capture lambda-s]
                     [make-aspect aspect])
         ;; aspect deployment
         deploy-fluid
         deploy-static
         deploy-top
         undeploy-top
         ;; pointcuts
         && || !
         top? top below above bottom bottom?
         target args some-args
         call? exec? call exec ;;LEVEL: no pc/adv
         cflowtop cflowbelow cflowbottom cflowabove
         cflow within
         ;; misc
         let let* letrec
         ;;LEVEL
         ^up ^down
         ;; internal
         app/weave lambda/aspects
         current-level
         w-c-m)

;; Join Point
;;                Proc           args
;; jp ::= call-jp a->b             a	;; procedure application ('a' can be values (ie. tuple {...})
;;     |  exec-jp a->b             a	;; procedure execution (cannot be advised only matched)
;;   AspectJ matches and transforms dispatches and calls, but not executions
;;   they're just poorly named (dispatch == `call', call == `execution')
;;     |  adv-jp  (a->b)->c->a->b  c	;; advice execution ... 'c' can be values as well

;; Pointcut
;; pc :: {[jp]*jp*[jp]}->c		;; above * jp * below

;; Advice
;; adv :: (a->b)->c->a->b

;; Aspect
;; creation: 
;; (aspect pc adv)
;; 
;; deployment:
;; deploy ::=    deploy-fluid  pc adv body  ;; dynamic scoping
;;         |     deploy-static pc adv body  ;; lexical scoping
;;         |     deploy-top    pc adv       ;; top-level scoping


;; aspect structure
(define-struct aspect (pc adv))

;; check used in deployment
(define-syntax-rule (check-aspect a)
  (when (not (aspect? a))
    (error "expecting an aspect constructed with (aspect <pc> <adv>), given " a)))

;; internal aspect structure
(define-struct asp (pc adv level)) ;; LEVEL

;; join point structures
(define-struct          jp (target args level)) ;;LEVEL
(define-struct (call-jp jp)(           ))
(define-struct (exec-jp jp)(           ))

;;; current execution level is a parameter
(define current-level (make-parameter 0)) ;; LEVEL

;;
;; COPIED FROM ASPECT-SCHEME.SS
;;
(define (c-c-m key) 
  (continuation-mark-set->list
   (current-continuation-marks) 
   key))

(define-syntax (w-c-m stx)
  (syntax-case stx ()
    [(_ tag mark body)
     (syntax ((lambda (x) x)
              (with-continuation-mark tag mark body)))]))

;;; level-shifting operators ;; LEVEL
;; up
(define-syntax ^up
  (lambda (stx)
    (syntax-case stx ()
      [(_ body ...)
       (syntax/loc stx (parameterize ((current-level (+ 1 (current-level))))
                         (begin body ...)))])))

;; down
(define-syntax ^down
  (lambda (stx)
    (syntax-case stx ()
      [(_ body ...)
       (syntax/loc stx (parameterize ((current-level (- (current-level) 1)))
                         (begin body ...)))])))

;; aspects are kept in parameters, according to their scope
(define dynamic-aspects  (make-parameter '()))
(define static-aspects   (make-parameter '()))
(define toplevel-aspects (make-parameter '()))

(define-struct undeployer (asp))

;; undeploys a top-level aspect
(define (undeploy-top ud)
  (toplevel-aspects (remove (undeployer-asp ud)
                            (toplevel-aspects)
                            eq?)))

;; deploy an aspect at the top level (global scope)
;; returns a handle that can be used to undeploy the aspect (with undeploy-top)
(define (deploy-top aspect)
  (check-aspect aspect)
  (let ((asp (make-asp (aspect-pc aspect) 
                       (aspect-adv aspect) 
                       (current-level))))
    (toplevel-aspects (cons asp (toplevel-aspects)))
    (make-undeployer asp)))

;; useful for testing
(define (clear-top)
  (toplevel-aspects '()))

;; scoped deployment
(define-syntaxes (deploy-fluid deploy-static)
  (let ([round (lambda (param)
                 (lambda (stx)
                   (syntax-case stx ()
                     [(_ aspect body ...)
                      (quasisyntax/loc stx 
                        (let ((a aspect))
                          (check-aspect a)
                          (parameterize ([#,param  
                                          (cons (make-asp (aspect-pc a) 
                                                          (aspect-adv a) 
                                                          (current-level))  
                                                (#,param))])
                            body ...)))])))])
    (values (round #`dynamic-aspects)   ;dynamically-scoped
            (round #`static-aspects)))) ;lexically-scoped


;; all functions capture and then restore their static aspects
(define-syntax lambda/aspects
  (lambda (stx)
    (syntax-case stx ()
      [(_ params body ...)
       (syntax/loc stx (let ([aspects (static-aspects)])
                         (lambda params
                           (parameterize ([static-aspects aspects])
                             body ...))))])))

;; level-capturing functions
(define-syntax lambda/aspects/capture
  (lambda (stx)
    (syntax-case stx ()
      [(_ params body ...)
       (syntax/loc stx (let ([aspects (static-aspects)]
                             [level (current-level)]) ; get current level
                         (lambda params
                           (parameterize ([static-aspects aspects]
                                          [current-level  level]) ; reinstate for body eval 
                             (begin body ...)))))])))

;; current aspects -- in descending order of application!
(define (current-aspects)
  (append (dynamic-aspects)
          (static-aspects)
          (toplevel-aspects)))


;; weaver

;; join points implemented as continuation marks

;; ensure that our continuation mark is unique and unforgeable
(define-struct jp-mark (id))
(define jp-mark-tag (make-jp-mark 'joinpoint))

(define (jp-context) 
  (continuation-mark-set->list
   (current-continuation-marks) 
   jp-mark-tag))

(define-syntax with-joinpoint
  (syntax-rules ()
    [(_ jp body ...)
     ((lambda (x) x)
      (with-continuation-mark jp-mark-tag jp
        (begin body ...)))]))

;; replacement for #%app
(define-syntax app/weave
  (lambda (stx)
    (syntax-case stx ()
      [(_ f a ...)
       (syntax/loc stx (app/weave/rt f a ...))])))

;; runtime weaving
(define (app/weave/rt fun-val . arg-vals)
  (let ([jp (make-call-jp fun-val arg-vals (current-level))] ;;LEVEL
        [jp+ (jp-context)]
        [proceed-level (current-level)])
    (with-joinpoint jp              
                    (let ((base-eval
                           (lambda arg-vals
                             (with-joinpoint (make-exec-jp fun-val arg-vals (current-level)) ;;LEVEL
                                             (parameterize ([current-level proceed-level])
                                               (apply fun-val arg-vals))))));; proceed at -1
                      (^up
                       (apply  (weave base-eval '() jp jp+ (current-aspects)) 
                               arg-vals))))))

;; build the composed advice chain
(define (weave fun-val jp- jp jp+ aspects)
  (foldr (lambda (aspect fun)
           (cond
             [(and (eq? (asp-level aspect) (jp-level jp))  
                   (with-joinpoint (make-call-jp (asp-pc aspect) 
                                                 (list jp- jp jp+) 
                                                 (current-level)) ;;LEVEL
                                   ((asp-pc aspect) jp- jp jp+)))
              => (lambda (ctxt-vals)
                   (lambda args
                     (with-joinpoint (make-call-jp (asp-adv aspect) 
                                                   ctxt-vals 
                                                   (current-level)) ;;LEVEL
                                     (apply (apply ((asp-adv aspect) fun) ctxt-vals) args))))] 
             [else fun]))                                                                       
         fun-val
         aspects))

;;;;;;;;;;;;;;;;;;;;;
;;; POINTCUTS ;;;;;;;
;;;;;;;;;;;;;;;;;;;;;

;; pointcuts -- strict combinators
(define ((&& . pcs) jp- jp jp+)
  (let loop ([pcs  pcs]
             [res  '()])
    (if (null? pcs)
        res
        (let ([r ((car pcs) jp- jp jp+)])
          (and r
               (loop (cdr pcs) (append res r)))))))

(define ((|| . pcs) jp- jp jp+)
  (let loop ([pcs pcs])
    (and (not (null? pcs))
         (or ((car pcs) jp- jp jp+)
             (loop (cdr pcs))))))

(define ((! pc) jp- jp jp+)
  (and (not (pc jp- jp jp+))
       '()))

;; pointcuts -- structural
(define (top? jp- jp jp+)
  (and (null? jp+)
       '()))

(define ((below pc) jp- jp jp+)
  (and (not (null? jp+))
       (pc (cons jp jp-) (car jp+) (cdr jp+))))

(define ((above pc) jp- jp jp+)
  (and (not (null? jp-))
       (pc (cdr jp-) (car jp-) (cons jp jp+))))

(define (bottom? jp- jp jp+)
  (and (null? jp-)
       '()))

;; pointcuts -- `binding'
(define (target jp- jp jp+)
  (list (jp-target jp)))

(define (args jp- jp jp+)
  (jp-args jp))

(define ((some-args as) jp- jp jp+)
  (foldl (lambda (a v l)
           (if a
               (cons v l)
               l))
         '()
         as
         (jp-args jp)))

;; pointcuts -- fundamental
(define ((true-pc) jp- jp jp+)
  '())

(define ((this-jp? p?) jp- jp jp+)
  (and (p? jp)
       '()))

(define-values (call? exec?)     ;; LEVEL: removed pc/adv
  (values (this-jp? call-jp?)
          (this-jp? exec-jp?)))

(define-values (call exec)       ;; LEVEL: removed pc/adv
  (let* ([(target= f) (this-jp? (lambda (jp)
                                  (equal? f (jp-target jp))))]
         [((this-target? p) f) (&& (target= f)
                                   p)])
    (values (this-target? call?)
            (this-target? exec?))))

(define (display-jps jp- jp jp+)
  (for-each (lambda (jp) (display `(< ,jp ,(jp-args jp)))) jp-)
  (display `(= ,jp ,(jp-args jp)))
  (for-each (lambda (jp) (display `(> ,jp ,(jp-args jp)))) jp+)
  (newline))

;; pointcuts - higher-order recursive
(define (((cflow-walk pc) step end) jp- jp jp+)
  ((|| pc
       (&& (! end)
           (step ((cflow-walk pc) step end)))) jp- jp jp+))

(define (top pc)
  (&& pc
      (! (cflowbelow pc))))

(define (bottom pc)
  (&& pc
      (! (cflowabove pc))))

;; pointcuts - cflow and co
(define (cflowtop pc)
  (cflowbelow (top pc)))

(define (cflowbelow pc)
  (below ((cflow-walk pc) below top?)))

(define (cflowabove pc)
  (above ((cflow-walk pc) above bottom)))

(define (cflowbottom pc)
  (cflowbelow (bottom pc)))

(define (cflow pc)
  ((cflow-walk pc) below top?))

(define (within f)
  (cflowbelow (&& (exec f)
                  (! (cflowabove call?)))))